structure Syntax :>
sig

    datatype typ = Nat | Arrow of typ * typ | Unit | Product of typ * typ | Sum of typ * typ;

    (* for this assignment, we represent names by a pair of a string and
       a unique int suffix (used for alpha-conversion).  
       freshName generates a new suffix that has never been used before,
       so freshName x is guaranteed to be different from all previously used variables.  

       Note that you can still use (x = y) to check if two names are equal.  

       In a term, variable names (x,i) are printed as x[i].  You can
       mostly ignore [i] when reading the term, and only look at them
       for disambiguation if things seem wrong.  

       *)
    type name = string * int
    val freshNameString : string -> name
    val freshName : name -> name
        
    datatype exp = 
	Var of name
      | Zero
      | Succ of exp
      | Rec of exp * exp * (name * name * exp)
      | Lam of typ * (name * exp)
      | App of exp * exp
      | Let of exp * (name * exp)
      | Triv 
      | Pair of exp * exp
      | ProjLeft of exp
      | ProjRight of exp
      | InLeft of typ * exp
      | InRight of typ * exp
      | Case of exp * (name * exp) * (name * exp)
        
    (* freshen the top-level bound variables in e if e is a Let/Lam/Rec *)
    val freshenTop : exp -> exp
        
    val subst : exp * (exp * name) -> exp

end =
struct

    datatype typ = Nat | Arrow of typ * typ | Unit | Product of typ * typ | Sum of typ * typ;

    type name = string * int

    local 
	val suffix = ref 0 
    in
        fun freshNameString (s : string) : name = 
            (suffix := (!suffix)+1;
             (s,(!suffix)))
            
	fun freshName ((s,i) : name) : name = 
	    freshNameString s
    end

    datatype exp = 
	Var of name
      | Zero
      | Succ of exp
      | Rec of exp * exp * (name * name * exp) 
      | Lam of typ * (name * exp)
      | App of exp * exp
      | Let of exp * (name * exp)
      | Triv 
      | Pair of exp * exp
      | ProjLeft of exp
      | ProjRight of exp
      | InLeft of typ * exp
      | InRight of typ * exp
      | Case of exp * (name * exp) * (name * exp)


    fun swapName (z : name, (x : name, y : name)) =
	if z = x then y
	else if z = y then x
	else z

    fun swapExp (e : exp, (z : name, w : name)) : exp =
	let (* give short names to the recursive call and 
               swapName call since they always get called with the same x and y *)
            fun se e = swapExp (e, (z, w))
	    fun sn x = swapName (x ,(z, w))
        in
            case e of 
                Var z => Var(sn z)
              | Zero => Zero
              | Succ e => Succ (se e)
              | Rec (e, e0, (x,y,e1)) => Rec (se e, se e0 , (sn x,sn y, se e1))
              | Lam (t, (x,e)) => Lam (t, (sn x, se e))
              | App (e1, e2) => App (se e1 , se e2)
              | Let(e1, (x, e2)) => Let(se e1, (sn x, se e2))
              | Triv => Triv
              | Pair(e1, e2) => Pair(se e1, se e2)
              | ProjLeft e => ProjLeft (se e)
              | ProjRight e => ProjRight (se e)
              | InLeft (t, e) => InLeft (t, se e)
              | InRight (t, e) => InRight (t, se e)
              | Case(e, (x1, e1), (x2, e2)) => Case(se e, (sn x1, se e1), (sn x2, se e2))
	end


    (* freshen the topmost binder in e *)
    fun freshenTop (e : exp) : exp =
        case e of 
            Let (e1, (x, e2)) => 
                let val x' = freshName x in
                    Let (e1, (x', swapExp (e2, (x, x'))))
                end
          | Rec (e, e0, (x,y,e1)) =>
                let val x' = freshName x
                    val y' = freshName y
                in Rec (e, e0 , (x',y', swapExp(swapExp (e1, (x,x')), (y,y'))))
                end
          | Lam (t, (x,e)) => 
                let val x' = freshName x
                in Lam (t,(x',swapExp(e, (x,x'))))
                end
          | Case (e, (x1,e1), (x2, e2)) =>
                let val x1' = freshName x1
                    val x2' = freshName x2
                in
                    Case (e, (x1', swapExp(e1, (x1,x1'))), (x2', swapExp(e2, (x2, x2'))))
                end
          | e => e

    fun subst (e : exp, (e' : exp , x' : name)) = 
	let (* give short names to the recursive call since they always get called with the same x and y *)
            fun se e = subst (e, (e', x'))
        in
            (* since we freshen the outermost bound variable here,
               the substitution will work even with shadowing and not capture
               *)
            case freshenTop e of
                Var x => if x = x' then e' else Var x
              | Zero => Zero
              | Succ e => Succ (se e)
              | Rec (e, e0, (x,y,e1)) => Rec (se e, se e0 , (x,y, se e1))
              | Lam (t, (x,e)) => Lam (t, (x, se e))
              | App (e1, e2) => App (se e1 , se e2)
              | Let(e1, (x, e2)) => Let (se e1, (x, se e2))
              | Triv => Triv
              | Pair(e1,e2) => Pair(se e1, se e2)
              | ProjLeft e => ProjLeft (se e)
              | ProjRight e => ProjRight (se e)
              | InLeft (t, e) => InLeft (t, se e)
              | InRight (t, e) => InRight (t, se e)
              | Case(e, (x1,e1), (x2,e2)) => Case(se e, (x1, se e1), (x2, se e2))
        end

end